/*
Copyright (C) 2004 Douglas Thain and the University of Notre Dame
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>

#include "chirp_client.h"
#include "chirp_protocol.h"

#include "timestamp.h"
#include "debug.h"
#include "auth_all.h"
#include "stringtools.h"
#include "xmalloc.h"
#include "list.h"

#ifdef HAS_LIBREADLINE
#include "readline/readline.h"
#include "readline/history.h"
#endif

#define BUFFER_SIZE 32768

static int do_open( const char *x, const char *y, const char *z );
static int do_close( const char *x, const char *y, const char *z );
static int do_get( const char *x, const char *y, const char *z );
static int do_put( const char *x, const char *y, const char *z );
static int do_cd( const char *x, const char *y, const char *z );
static int do_lcd( const char *x, const char *y, const char *z );
static int do_pwd( const char *x, const char *y, const char *z );
static int do_lpwd( const char *x, const char *y, const char *z );
static int do_getacl( const char *x, const char *y, const char *z );
static int do_setacl( const char *x, const char *y, const char *z );
static int do_ls( const char *x, const char *y, const char *z );
static int do_mv( const char *x, const char *y, const char *z );
static int do_rm( const char *x, const char *y, const char *z );
static int do_mkdir( const char *x, const char *y, const char *z );
static int do_rmdir( const char *x, const char *y, const char *z );
static int do_stat( const char *x, const char *y, const char *z );
static int do_timeout( const char *x, const char *y, const char *z );
static int do_debug( const char *x, const char *y, const char *z );
static int do_help( const char *x, const char *y, const char *z );
static int do_quit( const char *x, const char *y, const char *z );

static int do_put_recursive( const char *local_file, const char *remote_file );
static int do_get_recursive( const char *remote_file, const char *local_file );

static struct chirp_client *client=0;
static int timeout=3600;
static time_t stoptime=0;
static char current_host[CHIRP_PATH_MAX];
static char current_local_dir[CHIRP_PATH_MAX];
static char current_remote_dir[CHIRP_PATH_MAX];

static INT64_T xfer_items = 0;
static INT64_T xfer_bytes = 0;

struct command {
	char *name;
	int must_be_open;
	int minargs, maxargs;
	char *help;
	int (*handler) ( const char *x, const char *y, const char *z);
};

static struct command list[] =
{
{"open",    0, 1, 1, "<host>",                   do_open},
{"close",   1, 0, 0, "",                         do_close},
{"cd",      1, 1, 1, "<remotedir>",              do_cd},
{"lcd",     0, 1, 1, "<localdir>",               do_lcd},
{"pwd",     1, 0, 0, "",                         do_pwd},
{"lpwd",    0, 0, 0, "",                         do_lpwd},
{"get",	    1, 1, 2, "<remotefile> [localfile]", do_get},
{"put",	    1, 1, 2, "<localfile> [remotefile]", do_put},
{"getacl",  1, 0, 1, "[remotepath]",             do_getacl},
{"setacl",  1, 3, 3, "<remotepath> <user> <acl>",do_setacl},
{"ls",      1, 0, 2, "[-la] [remotepath]",       do_ls},
{"mv",      1, 2, 2, "<oldname> <newname>",      do_mv},
{"rm",	    1, 1, 1, "<file>",                   do_rm},
{"mkdir",   1, 1, 1, "<dir>",                    do_mkdir},
{"rmdir",   1, 1, 1, "<dir>",                    do_rmdir},
{"stat",    1, 1, 1, "<file>",                   do_stat},
{"timeout", 0, 1, 1, "<seconds>",                do_timeout},
{"debug",   0, 0, 1, "[subsystem]",              do_debug},
{"help",    0, 0, 0, "",                         do_help},
{"exit",    0, 0, 0, "",                         do_quit},
{"quit",    0, 0, 0, "",                         do_quit},
{0,0,0,0,0},
};

static void show_version( const char *cmd )
{
	printf("%s version %d.%d.%d built by %s@%s on %s at %s\n",cmd,CCTOOLS_VERSION_MAJOR,CCTOOLS_VERSION_MINOR,CCTOOLS_VERSION_MICRO,BUILD_USER,BUILD_HOST,__DATE__,__TIME__);
}

static void show_help( const char *cmd )
{
	printf("use: %s [options] [hostname]\n",cmd);
	printf("where options are:\n");
	printf(" -a <flag>  Require this authentication mode.\n");
	printf(" -d <flag>  Enable debugging for this subsystem.\n");
	printf(" -v         Show program version.\n");
	printf(" -h         This message.\n");
}
 

int main( int argc, char *argv[] )
{	
#ifdef HAS_LIBREADLINE
	char *temp;
#endif
	int did_explicit_auth = 0;
	int interactive_mode;
	char prompt[CHIRP_LINE_MAX];
	char line[CHIRP_LINE_MAX];
	char command[CHIRP_LINE_MAX];
	char arg1[CHIRP_LINE_MAX];
	char arg2[CHIRP_LINE_MAX];
	char arg3[CHIRP_LINE_MAX];
	char c;
	int fields;
	int i;

	debug_config(argv[0]);

	while((c=getopt(argc,argv,"a:d:vh"))!=(char)-1) {
		switch(c) {
			case 'a':
				auth_register_byname(optarg);
				did_explicit_auth = 1;
				break;
			case 'd':
				debug_flags_set(optarg);
				break;
			case 'v':
				show_version(argv[0]);
				exit(0);
				break;
			case 'h':
				show_help(argv[0]);
				exit(0);
				break;

		}
	}
 
	if(!did_explicit_auth) auth_register_all();
	interactive_mode = isatty(0);
	getcwd(current_local_dir,CHIRP_PATH_MAX);

	if(optind<argc) {
		stoptime = time(0) + timeout;
		if(do_open(argv[optind],0,0)<0) {
			fprintf(stderr,"couldn't open %s: %s\n",argv[optind],strerror(errno));
			return 1;
		}
	}

	while(1) {
		if(interactive_mode) {
			sprintf(prompt," chirp:%s:%s> ",current_host,current_remote_dir);
		} else {
			prompt[0] = 0;
		}

#ifdef HAS_LIBREADLINE
		temp = readline(prompt);		
		if(!temp) break;
		strcpy(line,temp);
		free(temp);
#else
		printf("%s",prompt);
		fflush(stdout);
		if(!fgets(line,CHIRP_LINE_MAX,stdin)) break;
#endif

		if(!line[0]) continue;

		add_history(line);		
		command[0] = 0;
		arg1[0] = 0;
		arg2[0] = 0;
		arg3[0] = 0;

		fields = sscanf(line,"%s %s %s %s",command,arg1,arg2,arg3);
		if(fields==0) continue;

		for(i=0;list[i].name;i++) {
			if(!strcmp(list[i].name,command)) {
				if(!client && list[i].must_be_open) {
					printf("not connected\n");
					break;
				} else if( ((fields-1)>=list[i].minargs) && ((fields-1)<=list[i].maxargs ) ) {
					stoptime = time(0)+timeout;
					if(list[i].handler(arg1,arg2,arg3)<0) {
						fprintf(stderr,"couldn't %s %s: %s\n",command,arg1,strerror(errno));
						if(!interactive_mode) return 1;
					}
					break;
				} else {
					printf("use: %s %s\n",command,list[i].help);
					break;
				}
			}
		}

		if(!list[i].name) {
			printf("unknown command: %s  (try 'help')\n",command);
		}
	}

	printf("\n");
	return 0;
}

static int do_open( const char *host, const char *y, const char *z )
{
	do_close(0,0,0);
	client = chirp_client_connect(host,1,stoptime);
	if(!client) {
		return -1;
	} else {
		printf("connected to %s\n",host);
		strcpy(current_host,host);
		current_remote_dir[0] = '/';
		return 0;
	}
}

static int do_close( const char *x, const char *y, const char *z )
{
	if(client) {
		chirp_client_disconnect(client);
		current_host[0] = 0;
		current_remote_dir[0] = 0;
		client = 0;
	}
	return 0;
}

static void complete_local_path( const char *file, char *full_path )
{
	char temp[CHIRP_PATH_MAX];
	if(file[0]!='/') {
		sprintf(temp,"%s/%s",current_local_dir,file);
	} else {
		strcpy(temp,file);
	}
	string_collapse_path(temp,full_path,1);
}

static void complete_remote_path( const char *file, char *full_path )
{
	char temp[CHIRP_PATH_MAX];
	if(file[0]!='/') {
		sprintf(temp,"%s/%s",current_remote_dir,file);
	} else {
		strcpy(temp,file);
	}
	string_collapse_path(temp,full_path,1);
}

static int do_cd( const char *file, const char *y, const char *z )
{
	char full_path[CHIRP_PATH_MAX];
	struct stat info;
	complete_remote_path(file,full_path);
	if(chirp_client_stat(client,full_path,&info,stoptime)<0) {
		return -1;
	} else {
		if(S_ISDIR(info.st_mode)) {
			string_collapse_path(full_path,current_remote_dir,1);
			return 0;
		} else {
			errno = ENOTDIR;
			return -1;
		}
	}
}

static int do_lcd( const char *file, const char *y, const char *z )
{
	char full_path[CHIRP_PATH_MAX];
	complete_local_path(file,full_path);
	if(chdir(full_path)==0) {
		getcwd(current_local_dir,CHIRP_PATH_MAX);
		return 0;
	} else {
		return -1;
	}
}

static int do_pwd( const char *x, const char *y, const char *z )
{
	printf("%s\n",current_remote_dir);
	return 0;
}

static int do_lpwd( const char *x, const char *y, const char *z )
{
	printf("%s\n",current_local_dir);
	return 0;
}

static int do_get_one_dir( const char *source_file, const char *target_file, int mode )
{
	char new_source_file[CHIRP_PATH_MAX];
	char new_target_file[CHIRP_PATH_MAX];
	struct list *work_list;
	const char *name;
	int result;

	work_list = list_create();

	result = mkdir(target_file,mode);
	if(result==0 || errno==EEXIST ) {
		result = chirp_client_opendir(client,source_file,stoptime);
		if(result==0) {
			while((name=chirp_client_readdir(client,stoptime))) {
				if(!strcmp(name,".")) continue;
				if(!strcmp(name,"..")) continue;
				list_push_tail(work_list,xstrdup(name));
			}
			while((name=list_pop_head(work_list))) {
				sprintf(new_source_file,"%s/%s",source_file,name);
				sprintf(new_target_file,"%s/%s",target_file,name);
				result = do_get_recursive(new_source_file,new_target_file);
				free((char*)name);
				if(result!=0) break;
			}
		} else {
			result = -1;
		}
	} else {
		result = -1;
	}

	while((name=list_pop_head(work_list))) free((char*)name);
	list_delete(work_list);

	return result;
}

static int do_get_one_link( const char *source_file, const char *target_file )
{
	char linkdata[CHIRP_PATH_MAX];
	int result;

	result = chirp_client_readlink(client,source_file,linkdata,sizeof(linkdata),stoptime);
	if(result>0) {
		linkdata[result] = 0;
		unlink(target_file);
		result = symlink(linkdata,target_file);
		if(result>=0) result = 0;
	}

	return result;
}

static int do_get_one_file( const char *source_file, const char *target_file, int mode, int length )
{
	FILE *file;
	timestamp_t start;
	double elapsed;
	int save_errno;
	int actual;

	file = fopen(target_file,"w");
	if(!file) {
		save_errno = errno;
		printf("\n");
		errno = save_errno;
		return -1;
	}

	fchmod(fileno(file),mode);

	start = timestamp_get();

	actual = chirp_client_getfile(client,source_file,file,stoptime);
	if(actual!=length) {
		save_errno = errno;
		fclose(file);
		printf("\n");
		errno = save_errno;
		return -1;
	}

	elapsed = (timestamp_get()-start)/1000000.0;

	if(length>=0) {
		fclose(file);
		printf("(%.2lf MB/s)\n",length/elapsed/1000000.0);
		return 0;
	} else {
		save_errno = errno;
		fclose(file);
		printf("\n");
		errno = save_errno;
		return -1;
	}
}

static int do_get_recursive( const char *source_file, const char *target_file )
{
	int result;
	struct stat info;

	result = chirp_client_lstat(client,source_file,&info,stoptime);
	if(result<0) {
		result = -1;
	} else {
		if(S_ISLNK(info.st_mode)) {
			printf("link %s\n->   %s\n",source_file,target_file);
			result = do_get_one_link(source_file,target_file);
		} else if(S_ISDIR(info.st_mode)) {
			printf("dir  %s\n->   %s\n",source_file,target_file);
			result = do_get_one_dir(source_file,target_file,info.st_mode);
		} else if(S_ISREG(info.st_mode)) {
			printf("file %s\n->   %s ",source_file,target_file);
			fflush(stdout);
			result = do_get_one_file(source_file,target_file,info.st_mode,info.st_size);
			xfer_bytes += info.st_size;
		} else {
			printf("???  %s\n->   %s\n",source_file,target_file);
			fflush(stdout);
			result = 0;
		}
		xfer_items++;
	}

	if(result<0) {
		printf("couldn't get %s: %s\n",source_file,strerror(errno));
	}

	return result;
}

static int do_get( const char *source_file, const char *target_file, const char *z )
{
	char target_full_path[CHIRP_PATH_MAX];
	char source_full_path[CHIRP_PATH_MAX];
	timestamp_t start, stop;
	double elapsed;
	int result;

	if(!target_file[0]) target_file=string_basename(source_file);

	complete_remote_path(source_file,source_full_path);
	complete_local_path(target_file,target_full_path);

	xfer_items = 0;
	xfer_bytes = 0;

	start = timestamp_get();
	result = do_get_recursive(source_full_path,target_full_path);
	stop = timestamp_get();

	elapsed = (stop-start)/1000000.0;

	printf("%lld files and %sB read in %.2fs ",
		xfer_items,
		string_metric(xfer_bytes,-1,0),
		elapsed);

	printf(" (%sB/s)\n",string_metric(xfer_bytes/elapsed,-1,0));

	return result;
}


static int do_put_one_dir( const char *source_file, const char *target_file, int mode )
{
	char new_source_file[CHIRP_PATH_MAX];
	char new_target_file[CHIRP_PATH_MAX];
	struct list *work_list;
	const char *name;
	int result;
	struct dirent *d;
	DIR *dir;

	work_list = list_create();

	result = chirp_client_mkdir(client,target_file,mode,stoptime);
	if(result==0 || errno==EEXIST ) {
		result = 0;
		dir = opendir(source_file);
		if(dir) {
			while((d=readdir(dir))) {
				if(!strcmp(d->d_name,".")) continue;
				if(!strcmp(d->d_name,"..")) continue;
				list_push_tail(work_list,xstrdup(d->d_name));
			}
			closedir(dir);
			while((name=list_pop_head(work_list))) {
				sprintf(new_source_file,"%s/%s",source_file,name);
				sprintf(new_target_file,"%s/%s",target_file,name);
				result = do_put_recursive(new_source_file,new_target_file);
				free((char*)name);
				if(result!=0) break;
			}
		} else {
			result = -1;
		}
	} else {
		result = -1;
	}

	while((name=list_pop_head(work_list))) free((char*)name);
	list_delete(work_list);

	return result;
}

static int do_put_one_link( const char *source_file, const char *target_file )
{
	char linkdata[CHIRP_PATH_MAX];
	int result;

	result = readlink(source_file,linkdata,sizeof(linkdata));
	if(result>0) {
		char linkdatafull[CHIRP_PATH_MAX];
		linkdata[result] = 0;
		complete_local_path(linkdata,linkdatafull);
		chirp_client_unlink(client,target_file,stoptime);
       		result = chirp_client_symlink(client,linkdatafull,target_file,stoptime);
		if(result>=0) result = 0;
	}

	return result;
}

static int do_put_one_file( const char *source_file, const char *target_file, int mode, int length )
{
	FILE *file;
	timestamp_t start;
	double elapsed;
	int save_errno;

	file = fopen(source_file,"r");
	if(!file) {
		save_errno = errno;
		printf("\n");
		errno = save_errno;
		return -1;
	}

	start = timestamp_get();

	length = chirp_client_putfile(client,target_file,file,mode,length,stoptime);
	if(length<0) {
		save_errno = errno;
		fclose(file);
		printf("\n");
		errno = save_errno;
		return -1;
	}

	elapsed = (timestamp_get()-start)/1000000.0;

	if(length>=0) {
		fclose(file);
		printf("(%sB/s)\n",string_metric(length/elapsed,-1,0));
		return 0;
	} else {
		save_errno = errno;
		fclose(file);
		printf("\n");
		errno = save_errno;
		return -1;
	}
}

static int do_put_recursive( const char *source_file, const char *target_file )
{
	int result;
	struct stat info;

	result = lstat(source_file,&info);
	if(result<0) {
		result = -1;
	} else {
		if(S_ISLNK(info.st_mode)) {
			printf("link %s\n->   %s\n",source_file,target_file);
			result = do_put_one_link(source_file,target_file);
		} else if(S_ISDIR(info.st_mode)) {
			printf("dir  %s\n->   %s\n",source_file,target_file);
			result = do_put_one_dir(source_file,target_file,info.st_mode);
		} else if(S_ISREG(info.st_mode)) {
			printf("file %s\n->   %s ",source_file,target_file);
			result = do_put_one_file(source_file,target_file,info.st_mode,info.st_size);
			xfer_bytes += info.st_mode;
		} else {
			printf("???  %s\n->   %s\n",source_file,target_file);
			result = 0;
		}
		xfer_items++;
	}

	if(result<0) {
		printf("couldn't put %s: %s\n",source_file,strerror(errno));
	}

	return result;
}

static int do_put( const char *source_file, const char *target_file, const char *z )
{
	char target_full_path[CHIRP_PATH_MAX];
	char source_full_path[CHIRP_PATH_MAX];
	timestamp_t start, stop;
	double elapsed;
	int result;

	if(!target_file[0]) target_file=string_basename(source_file);

	complete_local_path(source_file,source_full_path);
	complete_remote_path(target_file,target_full_path);

	xfer_items = 0;
	xfer_bytes = 0;

	start = timestamp_get();
	result = do_put_recursive(source_full_path,target_full_path);
	stop = timestamp_get();

	elapsed = (stop-start)/1000000.0;

	printf("%lld files and %sB written in %.2fs ",
		xfer_items,
		string_metric(xfer_bytes,-1,0),
		elapsed);

	printf(" (%sB/s)\n",string_metric(xfer_bytes/elapsed,-1,0));

	return result;
}

static int do_setacl( const char *file, const char *user, const char *acl )
{
	char full_path[CHIRP_PATH_MAX];
	complete_remote_path(file,full_path);
	return chirp_client_setacl(client,full_path,user,acl,stoptime);
}

static int do_getacl( const char *file, const char *y, const char *z )
{
	char full_path[CHIRP_PATH_MAX];
	const char *name;
	int result;

	if(!file[0]) file = ".";
	complete_remote_path(file,full_path);

	result = chirp_client_openacl(client,full_path,stoptime);
	if(result==0) {
		while((name=chirp_client_readacl(client,stoptime))) {
			printf("%s\n",name);
		}
	}

	return result;
}

static int print_long_list( const char *shortname, const char *longname )
{
	char linkname[CHIRP_PATH_MAX];
	struct stat info;
	int result;
	const char *type;

	result = chirp_client_lstat(client,longname,&info,stoptime);
	if(result<0) {
		printf("couldn't lstat %s: %s\n",shortname,strerror(errno));
		return -1;
	}

	linkname[0] = 0;

	if(S_ISLNK(info.st_mode)) {
		result = chirp_client_readlink(client,longname,linkname,sizeof(linkname),stoptime);
		if(result>0) {
			linkname[result] = 0;
			strcat(linkname,"\n");
		} else {
			strcpy(linkname,"???");
		}
		type = "link";
	} else if(S_ISDIR(info.st_mode)) {
		type = "dir ";
	} else if(S_ISREG(info.st_mode)) {
		type = "file";
	} else {
		type = "????";
	}

	printf("%4s %8lld %-40s %24s",type,(long long)info.st_size,shortname,linkname[0] ? linkname : ctime(&info.st_ctime));

	return 0;
}

static int do_ls( const char *options, const char *file, const char *z )
{
	char full_path[CHIRP_PATH_MAX];
	struct list *work_list;
	const char *name;
	int long_mode = 0;
	int all_mode = 0;
	int result;


	if(options[0]!='-') {
		file = options;
		options = "-";
	}

	options++;
	while(*options) {
		switch(*options) {
			case 'l':
				long_mode = 1;
				break;
			case 'a':
				all_mode = 1;
				break;
			default:
				printf("ls: unknown option: %c\n",*options);
				errno = EINVAL;
				return -1;
		}
		options++;
	}

	if(!file[0]) file = ".";
	complete_remote_path(file,full_path);

	work_list = list_create();

	result = chirp_client_opendir(client,full_path,stoptime);
	if(result==0) {
		while((name=chirp_client_readdir(client,stoptime))) {
			list_push_tail(work_list,xstrdup(name));
		}
		while((name=list_pop_head(work_list))) {
			char new_full_path[CHIRP_PATH_MAX];
			sprintf(new_full_path,"%s/%s",full_path,name);

			if(!all_mode && name[0]=='.') {
				/* do nothing */
			} else {
				if(long_mode) {
					print_long_list(name,new_full_path);
				} else {
					printf("%s\n",name);
				}
			}
			free((char*)name);
		}
	}

	while((name=list_pop_head(work_list))) free((char*)name);
	list_delete(work_list);

	return result;
}

static int do_rm_recursive( const char *file )
{
	struct list *work_list;
	const char *name;
	int result;

	work_list = list_create();

	result = chirp_client_unlink(client,file,stoptime);
	if(result==0 || errno==ENOENT) {
		result = 0;
	} else if(result<0 && errno==EISDIR) {
		result = chirp_client_opendir(client,file,stoptime);
		if(result==0) {
			while((name=chirp_client_readdir(client,stoptime))) {
				if(!strcmp(name,".")) continue;
				if(!strcmp(name,"..")) continue;
				list_push_tail(work_list,xstrdup(name));
			}
			while((name = list_pop_head(work_list))) {
				char newfile[CHIRP_PATH_MAX];
				sprintf(newfile,"%s/%s",file,name);
				result = do_rm_recursive(newfile);
				free((char*)name);
				if(result!=0) break;
			}
			result = chirp_client_rmdir(client,file,stoptime);
		} else {
			result = -1;
		}
	}

	if(result==0) {
		printf("removed: %s\n",file);
	} else {
		printf("couldn't rm %s: %s\n",file,strerror(errno));
	}

	while((name=list_pop_head(work_list))) free((char*)name);
	list_delete(work_list);

	return result;
}

static int do_rm( const char *file, const char *x, const char *z )
{
	char full_path[CHIRP_PATH_MAX];
	complete_remote_path(file,full_path);
	return do_rm_recursive(full_path);
}

static int do_rmdir( const char *file, const char *x, const char *z )
{
	char full_path[CHIRP_PATH_MAX];
	complete_remote_path(file,full_path);
	return chirp_client_rmdir(client,full_path,stoptime);
}

static int do_mkdir( const char *file, const char *x, const char *z )
{
	char full_path[CHIRP_PATH_MAX];
	int result;

	complete_remote_path(file,full_path);
	result = chirp_client_mkdir(client,full_path,0777,stoptime);

	if(result<0 && errno==EEXIST) result = 0;

	return result;
}

static int do_stat( const char *file, const char *x, const char *z )
{
	char full_path[CHIRP_PATH_MAX];
	struct stat info;

	complete_remote_path(file,full_path);

	if(chirp_client_stat(client,full_path,&info,stoptime)<0) {
		return -1;
	} else {

		/*
		Note that we deliberately cast call integers into long long
		because these types vary widely from platform to platform and
		may be 64 bit in the future.  This is overkill but foolproof.
		*/

		printf("device:  %lld\n",(long long)info.st_dev);
		printf("inode:   %lld\n",(long long)info.st_ino);
		printf("mode:    %04llo\n",(long long)info.st_mode);
		printf("nlink:   %lld\n",(long long)info.st_nlink);
		printf("uid:     %lld\n",(long long)info.st_uid);
		printf("gid:     %lld\n",(long long)info.st_gid);
		printf("rdevice: %lld\n",(long long)info.st_rdev);
		printf("size:    %lld\n",(long long)info.st_size);
		printf("blksize: %lld\n",(long long)info.st_blksize);
		printf("blocks:  %lld\n",(long long)info.st_blocks);
		printf("atime:   %s",ctime(&(info.st_atime)));
		printf("mtime:   %s",ctime(&(info.st_mtime)));
		printf("ctime:   %s",ctime(&(info.st_ctime)));

		return 0;
	}
 }

static int do_mv( const char *old_file, const char *new_file, const char *z )
{
	char old_full_path[CHIRP_PATH_MAX];
	char new_full_path[CHIRP_PATH_MAX];
	complete_remote_path(old_file,old_full_path);
	complete_remote_path(new_file,new_full_path);
	return chirp_client_rename(client,old_full_path,new_full_path,stoptime);
}

static int do_debug( const char *flag, const char *y, const char *z )
{
	if(flag[0]) {
		if(debug_flags_set(flag)) {
			return 0;
		} else {
			printf("Valid debugging flags are: ");
			debug_flags_print(stdout);
			printf("\n");
			errno = EINVAL;
			return -1;
		}
	} else {
		debug_flags_clear();
		return 0;
	}	
	
}

static int do_timeout( const char *x, const char *y, const char *z )
{
	timeout = atoi(x);
	printf("Timeout is %d seconds.\n",timeout);
	return 0;
}

static int do_quit( const char *x, const char *y, const char *z )
{
	exit(0);
	return -1;
}

static int do_help( const char *x, const char *y, const char *z )
{
	int i;
	printf("Commands:\n");
	for(i=0;list[i].name;i++) {
		printf("  %8s %s\n",list[i].name,list[i].help);
	}
	printf("\nDebugging subsystems are:\n");
	debug_flags_print(stdout);
	printf("\n");
	return 0;
}

