/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include "pfs_channel.h"
#include "pfs_process.h"
#include "pfs_dispatch.h"
#include "pfs_poll.h"
#include "pfs_service.h"
#include "pfs_critical.h"
#include "pfs_control.h"

extern "C" {
#include "tracer.h"
#include "stringtools.h"
#include "auth_all.h"
#include "xmalloc.h"
#include "create_dir.h"
#include "file_cache.h"
#include "debug.h"
#include "getopt.h"
#include "chirp_resolve.h"
}

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <limits.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <pwd.h>

pid_t trace_this_pid = -1;

int pfs_master_timeout = 60;
int pfs_force_keepalive = 0;
struct file_cache *pfs_file_cache = 0;
int pfs_force_stream = 0;
int pfs_force_cache = 0;
int pfs_sloppy_cache = 0;
char *pfs_false_uname = 0;
uid_t pfs_uid = 0;
gid_t pfs_gid = 0;

long long pfs_syscall_count = 0;
long long pfs_read_count = 0;
long long pfs_write_count = 0;

/*
This process at the very top of the traced tree
and its final exit status, which we use to determine
our own exit status
*/

static pid_t root_pid = -1;
static int root_exitstatus = 0;
static int channel_size = 10;

extern int pfs_service_http_proxy_config( const char *str );

static void show_version( const char *cmd )
{
	printf("%s version %d.%d.%d built by %s@%s on %s at %s\n",cmd,CCTOOLS_VERSION_MAJOR,CCTOOLS_VERSION_MINOR,CCTOOLS_VERSION_MICRO,BUILD_USER,BUILD_HOST,__DATE__,__TIME__);
	exit(1);
}

static void print_filesystem_message( const char *name )
{
	if(pfs_service_lookup(name)) {
		printf("    /%-10s     (enabled)\n",name);
	} else {
		printf("    /%-10s     *DISABLED*\n",name);
	}
}

static void show_use( const char *cmd )
{
		/* 80-column text marker */
                /******************************************************************************/
	printf("Use: %s [options] <command> ...\n",cmd);
	printf("Where options and environment variables are:\n");
	printf("  -a <list>  Use these Chirp authentication methods.   (PARROT_CHIRP_AUTH)\n");
	printf("  -b <bytes> Set the I/O block size hint.              (PARROT_BLOCK_SIZE)\n");
	printf("  -c <host>  Resolve filenames via a Chirp server   (PARROT_RESOLVE_CHIRP)\n");
	printf("  -d <name>  Enable debugging for this sub-system.    (PARROT_DEBUG_FLAGS)\n");
	printf("  -F         Force all I/O protocols to use the cache.(PARROT_FORCE_CACHE)\n");
	printf("  -G <num>   Fake this gid; Real gid stays the same.          (PARROT_GID)\n");
	printf("  -h         Show this screen.\n");
	printf("  -k         Enable keepalive on all app sockets.       (PARROT_KEEPALIVE)\n");
	printf("  -m <file>  Use this file as a mountlist.             (PARROT_MOUNT_FILE)\n");
	printf("  -M/foo=/bar Mount (redirect) /foo to /bar.         (PARROT_MOUNT_STRING)\n");
	printf("  -N <name>  Pretend that this is my hostname.          (PARROT_HOST_NAME)\n");
	printf("  -o <file>  Send debugging messages to this file.     (PARROT_DEBUG_FILE)\n");
	printf("  -O <bytes> Rotate debug files of this size.     (PARROT_DEBUG_FILE_SIZE)\n");
	printf("  -p <host>  Use this proxy for HTTP requests.         (PARROT_HTTP_PROXY)\n");
	printf("  -s         Use streaming protocols without caching.(PARROT_FORCE_STREAM)\n");
	printf("  -S         Sloppy caching                          (PARROT_SLOPPY_CACHE)\n");
	printf("  -t <dir>   Where to store temporary files.             (PARROT_TEMP_DIR)\n");
	printf("  -T <time>  Maximum amount of time to retry failures.    (PARROT_TIMEOUT)\n");
	printf("  -U <num>   Fake this uid; Real uid stays the same.          (PARROT_UID)\n");
	printf("  -v         Display version number.\n");
	printf("\n");
	printf("Known debugging sub-systems are: ");
	debug_flags_print(stdout);
	printf("\n\n");
	printf("The argument to -T is assumed to be in seconds unless followed\n");
	printf("by an 's' 'm' 'h' or 'd' to indicate seconds, minutes, hours, or days.\n");
	printf("For example: parrot -T 5h tcsh\n");
	printf("\n");
	printf("Known filesystem types are:\n");
	print_filesystem_message("local");
	print_filesystem_message("http");
	print_filesystem_message("ftp");
	print_filesystem_message("anonftp");
	print_filesystem_message("gsiftp");
	print_filesystem_message("nest");
	print_filesystem_message("chirp");
	print_filesystem_message("rfio");
	print_filesystem_message("dcap");
	printf("\n");
	exit(1);
}

static int parse_time_string( const char *str )
{
	int value;
	char mod;

	if(sscanf(str,"%d%c",&value,&mod)==2) {
		switch(mod) {
			case 's': return value;
			case 'm': return value*60;
			case 'h': return value*60*60;
			case 'd': return value*60*60*24;
		}
	} else if(sscanf(str,"%d",&value)==1) {
		return value;
	}

	fprintf(stderr,"Sorry, I don't understand the time value '%s'\n",str);
	exit(1);
}

/*
For all of the signals that we handle, we want to
run the handler without interruption from other signals.
*/

static void install_handler( int sig, void (*handler)(int sig))
{
	struct sigaction s;

	s.sa_handler = handler;
	sigfillset(&s.sa_mask);
	s.sa_flags = 0; 

	sigaction(sig,&s,0);
}

/*
It would be nice if we could clean up everyone quietly and then
some time later, kill hard.  However, on Linux, if someone kills
us before we have a chance to clean up, then due to a "feature"
of ptrace, all our children will be left stuck in a debug-wait
state.  So, rather than chance ourselves getting killed, we
will be very agressive about cleaning up.  Upon receiving any
shutdown signal, we immediately blow away everyone involved,
and then kill ourselves.
*/

static void kill_everyone( int sig )
{
	debug(D_NOTICE,"received signal %d (%s), killing all my children...",sig,string_signal(sig));
	pfs_process_killall();
	debug(D_NOTICE,"sending myself %d (%s), goodbye!",sig,string_signal(sig));
	while(1) {
		signal(sig,SIG_DFL);
		sigsetmask(~sigmask(sig));
		kill(getpid(),sig);
		kill(getpid(),SIGKILL);
	}
}

/*
Other less deadly signals we simply pass through to the root
of our children for its consideration.
*/

static void pass_through( int sig )
{
	pfs_process_raise(root_pid,sig,1);
}

/*
We will be fighting with our child processes for control of 
the terminal.  Whenever a we want to write to the terminal,
we will get a SIGTTOU or SIGTTIN instructing us that its
not our turn.  But, we are in charge!  So, upon receipt
of these signals, grab control of the terminal.
*/

static void control_terminal( int sig )
{
	if(sig==SIGTTOU) {
		tcsetpgrp(1,getpid());
		tcsetpgrp(2,getpid());
	} else {
		tcsetpgrp(0,getpid());
	}
}

/*
Here is the meat and potatoes.  We have discovered that
something interesting has happened to this pid. Decode
the event and take the appropriate action.
*/

static void handle_event( pid_t pid, int status, struct rusage usage )
{
	struct pfs_process *p;
	int signum;

	p = pfs_process_lookup(pid);
	if(!p) {
		debug(D_PROCESS,"killing unexpected pid %d",pid);
		kill(pid,SIGKILL);
		return;
	}

	if(WIFEXITED(status)) {
		debug(D_PROCESS,"pid %d exited normally with code %d",pid,WEXITSTATUS(status));
		pfs_process_stop(p,status,usage);
		if(pid==root_pid) root_exitstatus = status;
	} else if(WIFSIGNALED(status)) {
		signum = WTERMSIG(status);
		debug(D_PROCESS,"pid %d exited abnormally with signal %d (%s)",pid,signum,string_signal(signum));
		pfs_process_stop(p,status,usage);
		if(pid==root_pid) root_exitstatus = status;
	} else if(WIFSTOPPED(status)) {
		signum = WSTOPSIG(status);
		if(signum==SIGTRAP) {
			pfs_dispatch(p,0);
		} else {
			debug(D_PROCESS,"pid %d received signal %d (%s) (state %d)",pid,signum,string_signal(signum),p->state);
			if(signum==SIGTTIN) {
				tcsetpgrp(0,pid);
				tracer_continue(p->tracer,SIGCONT);
			} else if(signum==SIGTTOU) {
				tcsetpgrp(1,pid);
				tcsetpgrp(2,pid);
				tracer_continue(p->tracer,SIGCONT);
			} else {
				tracer_continue(p->tracer,signum);
			}
		}
	} else {
		fatal("pid %d stopped with strange status %d",pid,status);
	}
}

struct timeval clock_to_timeval( clock_t c )
{
	struct timeval result;
	result.tv_sec = c/CLOCKS_PER_SEC;
	result.tv_usec = (c - result.tv_sec)*1000000/CLOCKS_PER_SEC;
	return result;
}

static void handle_sigchld( int sig  )
{
	pfs_poll_abort();
}

static void handle_sigio( int sig )
{
	pfs_process_sigio();
}

int main( int argc, char *argv[] )
{
	pid_t pid=0;
	int signum;
	int status;
	struct pfs_process *p;
	char *s;
	int i;
	int chose_auth=0;
	char temp_dir[PFS_PATH_MAX];
	struct passwd *pw;
	struct rusage usage;
	char c;

	debug_config(argv[0]);

	if(getenv("PARROT_ENABLED")) {
		fprintf(stderr,"sorry, parrot cannot be run inside of itself.\n");
		exit(1);
	}

	debug_config_fatal(pfs_process_killall);
	debug_config_getpid(pfs_process_getpid);

	install_handler(SIGQUIT,kill_everyone);
	install_handler(SIGILL,kill_everyone);
	install_handler(SIGABRT,kill_everyone);
	install_handler(SIGIOT,kill_everyone);
	install_handler(SIGBUS,kill_everyone);
	install_handler(SIGFPE,kill_everyone);
	install_handler(SIGSEGV,kill_everyone);
	install_handler(SIGTERM,kill_everyone);
	install_handler(SIGHUP,pass_through);
	install_handler(SIGINT,pass_through);
	install_handler(SIGTTIN,control_terminal);
	install_handler(SIGTTOU,control_terminal);
	install_handler(SIGCHLD,handle_sigchld);
	install_handler(SIGIO,handle_sigio);

	pfs_uid = getuid();
	pfs_gid = getgid();

	putenv("PARROT_ENABLED=TRUE");

	s = getenv("PARROT_RESOLVE_CHIRP");
	if(s) chirp_resolve_chirp_config(s);

	s = getenv("PARROT_BLOCK_SIZE");
	if(s) pfs_service_set_block_size(atoi(s));

	s = getenv("PARROT_HTTP_PROXY");
	if(s) pfs_service_http_proxy_config(s);

	s = getenv("PARROT_MOUNT_FILE");
	if(s) chirp_resolve_file_config(s);

	s = getenv("PARROT_MOUNT_STRING");
	if(s) chirp_resolve_manual_config(s);

	s = getenv("PARROT_FORCE_STREAM");
	if(s) pfs_force_stream = 1;

	s = getenv("PARROT_FORCE_CACHE");
	if(s) pfs_force_cache = 1;

	s = getenv("PARROT_SLOPPY_CACHE");
	if(s) pfs_sloppy_cache = 1;

	s = getenv("PARROT_HOST_NAME");
	if(s) pfs_false_uname = s;

	s = getenv("PARROT_UID");
	if(s) pfs_uid = atoi(s);

	s = getenv("PARROT_GID");
	if(s) pfs_gid = atoi(s);

	s = getenv("PARROT_TIMEOUT");
	if(s) pfs_master_timeout = parse_time_string(s);

	s = getenv("PARROT_KEEPALIVE");
	if(s) pfs_force_keepalive = 1;

	s = getenv("PARROT_DEBUG_FLAGS");
	if(s) {
		char *x = xstrdup(s);
		int nargs;
		char **args;
		if(string_split(x,&nargs,&args)) {
			for(int i=0;i<nargs;i++) {
				debug_flags_set(args[i]);
			}
		}
		free(x);
	}

	s = getenv("PARROT_CHIRP_AUTH");
	if(s) {
		char *x = xstrdup(s);
		int nargs;
		char **args;
		if(string_split(x,&nargs,&args)) {
			for(int i=0;i<nargs;i++) {
				auth_register_byname(args[i]);
				chose_auth = 1;
			}
		}
		free(x);
	}

	pw = getpwuid(getuid());
	if(pw) {
		sprintf(temp_dir,"%s/.parrot/cache",pw->pw_dir);
	} else {
		sprintf(temp_dir,"/tmp/parrot");
	}

	while((c=getopt(argc,argv,"+ha:b:B:C:d:FG:km:M:N:o:O:p:sSt:T:U:v"))!=(char)-1) {
		switch(c) {
		case 'a':
			if(!auth_register_byname(optarg)) {
				fprintf(stderr,"unknown auth type: %s\n",optarg);
				return 1;
			}
			chose_auth = 1;
			break;
		case 'b':
			pfs_service_set_block_size(atoi(optarg));
			break;
		case 'B':
			pfs_service_set_block_size(atoi(optarg));
			break;
		case 'C':
			chirp_resolve_chirp_config(optarg);
			break;
		case 'd':
			if(!debug_flags_set(optarg)) show_use(argv[0]);
			break;
		case 'F':
			pfs_force_cache = 1;
			break;	
		case 'G':
			pfs_gid = atoi(optarg);
			break;
		case 'k':
			pfs_force_keepalive = 1;
			break;
		case 'm':
			chirp_resolve_file_config(optarg);
			break;
		case 'M':
			chirp_resolve_manual_config(optarg);
			break;
		case 'N':
			pfs_false_uname = optarg;
			break;
		case 'o':
			debug_config_file(optarg);
			break;
		case 'O':
			debug_config_file_size(atoi(optarg));
			break;
		case 'p':
			pfs_service_http_proxy_config(optarg);
			break;
		case 's':
			pfs_force_stream = 1;
			break;
		case 'S':
			pfs_sloppy_cache = 1;
			break;
		case 't':
			strcpy(temp_dir,optarg);
			break;
		case 'T':
			pfs_master_timeout = parse_time_string(optarg);
			break;
		case 'U':
			pfs_uid = atoi(optarg);
			break;
		case 'v':
			show_version(argv[0]);
			break;
		default:
			show_use(argv[0]);
			break;
	}
	}

	if(optind>=argc) show_use(argv[0]);

	pfs_file_cache = file_cache_init(temp_dir);
	if(!pfs_file_cache) fatal("couldn't setup cache in %s: %s\n",temp_dir,strerror(errno));

	if(!chose_auth) auth_register_all();

	if(!pfs_channel_init(channel_size*1024*1024)) fatal("couldn't establish I/O channel");	

	pfs_poll_init();

	CRITICAL_BEGIN

	if(pid==0) {
		pid = fork();
		if(pid>0) {
			debug(D_PROCESS,"pid %d started",pid);
		} else if(pid==0) {
			for(i=0;i<sysconf(_SC_OPEN_MAX);i++) {
				if(i!=pfs_channel_fd()) close(i);
			}
			setpgrp();
			tracer_prepare();
			kill(getpid(),SIGSTOP);
			execvp(argv[optind],&argv[optind]);
			debug(D_NOTICE,"unable to execute %s: %s",argv[optind],strerror(errno));
			_exit(1);
		} else {
			debug(D_NOTICE,"unable to fork %s: %s",argv[optind],strerror(errno));
			exit(1);
		}
	}

	root_pid = pid;
	debug(D_PROCESS,"attaching to pid %d",pid);
	p = pfs_process_create(pid,getpid(),0,SIGCHLD);
	if(!p) {
		kill(pid,SIGKILL);
		fatal("unable to attach to pid %d: %s",pid,strerror(errno));
	}

	p->state = PFS_PROCESS_STATE_USER;
	strcpy(p->name,argv[optind]);

	while(pfs_process_count()>0) {
		while(1) {
			int flags;
			if(trace_this_pid!=-1) {
				flags = WUNTRACED|__WALL;
			} else {
				flags = WUNTRACED|__WALL|WNOHANG;
			}
			pid = wait4(trace_this_pid,&status,flags,&usage);
			if(pid>0) {
				handle_event(pid,status,usage);
				continue;
			} else {
				break;
			}
		}
		if(pfs_process_count()>0) pfs_poll_sleep();
	}

	debug(D_SUMMARY,"%lld syscalls",pfs_syscall_count);
	debug(D_SUMMARY,"%lld bytes read",pfs_read_count);
	debug(D_SUMMARY,"%lld bytes written",pfs_write_count);

	if(WIFEXITED(root_exitstatus)) {
		status = WEXITSTATUS(root_exitstatus);
		debug(D_PROCESS,"%s exited normally with status %d",argv[optind],status);
		return status;
	} else {
		signum = WTERMSIG(root_exitstatus);
		debug(D_PROCESS,"%s exited abnormally with signal %d (%s)",argv[optind],signum,string_signal(signum));
		return 1;
	}
}

