/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include "pfs_service.h"

extern "C" {
#include "debug.h"
#include "stringtools.h"
#include "domain_name.h"
#include "link.h"
#include "file_cache.h"
#include "full_io.h"
}

#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/statfs.h>

#define HTTP_LINE_MAX 4096
#define HTTP_PORT 80
#define HTTP_FILE_MODE (S_IFREG | 0444)
#define HTTP_TIMEOUT 3600

static char proxy_host[256] = "";
static int proxy_port = HTTP_PORT;

int pfs_service_http_proxy_config( const char *str )
{
	int fields;
	fields = sscanf(str,"%[^:]:%d",proxy_host,&proxy_port);
	if(fields==1 || fields==2) {
		return 1;
	} else {
		return 0;
	}
}

static int http_response_to_errno( int response ) 
{
	if(response<=299) {
		return 0;
	} else if(response<=399) {
		return EBUSY;
	} else if(response==400) {
		return EINVAL;
	} else if(response<=403) {
		return EACCES;
	} else if(response==404) {
		return ENOENT;
	} else if(response<=406) {
		return EINVAL;
	} else if(response==407) {
		return EACCES;
	} else if(response==408) {
		return ETIMEDOUT;
	} else if(response<=410) {
		return ENOENT;
	} else if(errno<=499) {
		return EINVAL;
	} else {
		return EIO;
	}
}

static struct link * http_fetch( pfs_name *name, const char *action )
{
	char line[HTTP_LINE_MAX];
	char url[HTTP_LINE_MAX];
	char addr[LINK_ADDRESS_MAX];
	struct link *link;
	int save_errno;
	int response;
	char actual_host[HTTP_LINE_MAX];
	int actual_port;

	if(!name->host[0]) {
		errno = ENOENT;
		return 0;
	}

	if(proxy_host[0]) {
		strcpy(actual_host,proxy_host);
		actual_port = proxy_port;
	} else {
		strcpy(actual_host,name->host);
		if(name->port==0) {
			actual_port = HTTP_PORT;
		} else {
			actual_port = name->port;
		}
	}

	sprintf(url,"http://%s%s",name->host,name->rest);

	reload:	

	debug(D_HTTP,"connect %s port %d",actual_host,actual_port);
	if(!domain_name_lookup(actual_host,addr)) return 0;

	link = link_connect(addr,actual_port,LINK_FOREVER);
	if(!link) return 0;

	sprintf(line,"%s %s HTTP/1.0\n\n",action,url);
	debug(D_HTTP,"%s",line);
	link_write(link,line,strlen(line),LINK_FOREVER);

	if(link_readline(link,line,HTTP_LINE_MAX,LINK_FOREVER)) {
		string_chomp(line);
		debug(D_HTTP,"%s",line);	
		if(sscanf(line,"HTTP/%*d.%*d %d",&response)==1) {
			url[0] = 0;
			while(link_readline(link,line,HTTP_LINE_MAX,LINK_FOREVER)) {
				string_chomp(line);
				debug(D_HTTP,"%s",line);
				sscanf(line,"Location: %s",url);
				if(strlen(line)<=2) {
					break;
				}
			}

			if(response==200) {
				return link;
			} else if(response==301) {
				link_close(link);
				if(url[0]) {
					goto reload;
				} else {
					errno = ENOENT;
					return 0;
				}
			} else {
				link_close(link);
				errno = http_response_to_errno(response);
				return 0;
			}
		} else {
			debug(D_HTTP,"malformed response");
			save_errno = EIO;
		}
	} else {
		debug(D_HTTP,"malformed response");
		save_errno = EIO;
	}

	link_close(link);
	errno = save_errno;
	return 0;
}


class pfs_file_http : public pfs_file
{
private:
	struct link *link;

public:
	pfs_file_http( pfs_name *n, struct link *l ) : pfs_file(n) {
		link = l;
	}

	virtual int close() {
		link_close(link);
		return 0;
	}

	virtual pfs_ssize_t read( void *d, pfs_size_t length, pfs_off_t offset ) {
		return link_read(link,(char*)d,length,LINK_FOREVER);
	}
};

class pfs_service_http : public pfs_service {
public:
	virtual int get_default_port() {
		return HTTP_PORT;
	}

	virtual pfs_file * open( pfs_name *name, int flags, mode_t mode ) {
		struct link *link;

		if((flags&O_ACCMODE)!=O_RDONLY) {
			errno = EROFS;
			return 0;
		}
 
		link = http_fetch(name,"GET");
		if(link) {
			return new pfs_file_http(name,link);
		} else {
			return 0;
		}
	}

	virtual int stat( pfs_name *name, struct stat *buf ) {
		struct link *link;
		link = http_fetch(name,"HEAD");
		if(link) {
			link_close(link);
			pfs_service_emulate_stat(name,buf);
			buf->st_mode = HTTP_FILE_MODE;
			return 0;
		} else {
			return -1;
		}	
	}

	virtual int lstat( pfs_name *name, struct stat *buf ) {
		return this->stat(name,buf);
	}
};

static pfs_service_http pfs_service_http_instance;
pfs_service *pfs_service_http = &pfs_service_http_instance;


